/**
 * Hunter Adams (vha3@cornell.edu)
 * Bruce Land (brl4@cornell.edu)
 * 
 * HARDWARE CONNECTIONS
 *  - GPIO 16 ---> VGA Hsync
 *  - GPIO 17 ---> VGA Vsync
 *  - GPIO 18 ---> 330 ohm resistor ---> VGA Green lo-bit |__ both wired to 150 ohm to ground 
 *  - GPIO 19 ---> 220 ohm resistor ---> VGA Green hi_bit |   and to VGA Green
 *  - GPIO 20 ---> 330 ohm resistor ---> VGA Blue
 *  - GPIO 21 ---> 330 ohm resistor ---> VGA Red
 *  - RP2040 GND ---> VGA GND
 *
 * RESOURCES USED
 *  - PIO state machines 0, 1, and 2 on PIO instance 0
 *  - DMA channels 0, 1, 2, and 3
 *  - 153.6 kBytes of RAM (for pixel color data)
 *
 * Protothreads v1.1.1
 * Threads:
 * ------------------
 * core 0:
 * Lattice Boltzmann -- control
 * blink LED25 
 * ------------------
 * core 1:
 * Lattice Boltzmann -- coprocessor
 * Serial i/o 
 *///
// ==========================================
// === VGA graphics library
// ==========================================
#include "vga16_graphics.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "pico/stdlib.h"
#include "hardware/pio.h"
#include "hardware/dma.h"
#include "pico/float.h"
#include "pico/double.h"

// ==========================================
// === protothreads globals
// ==========================================
#include "hardware/sync.h"
#include "hardware/timer.h"
#include "pico/multicore.h"
#include "hardware/clocks.h"
#include "string.h"
// protothreads header
#include "pt_cornell_rp2040_v1_3.h"

// (uses low level hardware spin lock) 
// used to sync the two solvers on two cores
struct pt_sem collide_go_s, collide_done_s ;
struct pt_sem stream_go_s, stream_done_s ;

// ==================================================
// === Lattice Boltzmann code from Daniel V. Schroeder
// ==================================================
/*
A lattice-Boltzmann fluid simulation in JavaScript, using HTML5 canvas for graphics	
	Copyright 2013, Daniel V. Schroeder
  >>> Modifed for C on  Pi Pico by Bruce Land 2022 <<<
	Permission is hereby granted, free of charge, to any person obtaining a copy of 
	this software and associated data and documentation (the "Software"), to deal in 
	the Software without restriction, including without limitation the rights to 
	use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies 
	of the Software, and to permit persons to whom the Software is furnished to do 
	so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included in all 
	copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
	INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A 
	PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR 
	ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
	OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
	OTHER DEALINGS IN THE SOFTWARE.

	Except as contained in this notice, the name of the author shall not be used in 
	advertising or otherwise to promote the sale, use or other dealings in this 
	Software without prior written authorization.
  */
								
	#define xdim  200			// grid dimensions for simulation 70x30 works also 200x40
	#define ydim 40
	
	char tracerCheck = true ;
	char running = true;						// will be true when running
	int stepCount = 0;
	int startTime = 0;
	float four9ths = (4.0f / 9.0f) ;					// abbreviations
	float one9th = (1.0f / 9.0f) ;
	float one36th = (1.0f / 36.0f) ;
    float one = (1.0f) ;
    float half = (0.5f) ;
	//float rho_twiddle = (0.97) ;
	int barrierCount = 0;
	int barrierxSum = 0;
	int barrierySum = 0;
	
	// Create the arrays of fluid particle densities, etc. (using 1D arrays for speed):
	// To index into these arrays, use x + y*xdim, traversing rows first and then columns.
	float n0 [xdim*ydim];			// microscopic densities along each lattice direction
	float nN [xdim*ydim];
	float nS [xdim*ydim];
	float nE [xdim*ydim];
	float nW [xdim*ydim];
	float nNE [xdim*ydim];
	float nSE [xdim*ydim];
	float nNW [xdim*ydim];
	float nSW [xdim*ydim];
	float rho [xdim*ydim];			// macroscopic density
	float ux  [xdim*ydim];			// macroscopic x-velocity
	float uy  [xdim*ydim];          // macroscopic y-velocity
	//Fix curl  [xdim*ydim];
	char barrier  [xdim*ydim];		// boolean array of barrier locations

    // Initialize tracers (but don't place them yet):
    #define n_trace  200
    #define tracer_enable 1
    float trace_x [n_trace];
    float trace_y [n_trace];

  
  // unstable above 0.10 or so
  float fluid_speed = (0.10f) ;
  // 25 bit is stable to viscosity 0.005 at speed 0.15
  float fluid_viscosity = (0.007f) ;
 // #define stepsPerFrame 1

// =========================================
// LB init function
// =========================================
void init_LB(void){
	// Initialize to a steady rightward flow with no barriers:
	for (int y=0; y<ydim; y++) {
		for (int x=0; x<xdim; x++) {
			barrier[x+y*xdim] = false;
		}
	}
} // end init_LB

void init_trace(void){
	int i;
	for( i=0; i<n_trace; i++){
		trace_x[i] = (2.0f + 5.0f*(i/ydim)) ;
		trace_y[i] = (4.0f + (i % (ydim-6))) ;
	}
}

void move_trace(void){
	int i, index;
	for(i=0; i<n_trace; i++){
		int index = (int)(trace_x[i] + 0.5f) + xdim*(int)(trace_y[i] + 0.5f) ;
		trace_x[i] += ux[index] ;
		//if(Fix2i(trace_x[i])<1) 
		trace_y[i] += uy[index] ;
		if( (int)(trace_x[i])>xdim-2) {
			trace_x[i] = (2.0f)  ;
            //trace_y[i] = i2Fix14(4 +(i % (ydim-6))) ;  ;
			trace_y[i] = (4.0f + (rand() % (ydim-6))) ;  ;
		}
		//if(Fix2i(trace_y[i])<2 || Fix2i(trace_y[i])>ydim-2) trace_y[i] = i2Fix(10+(rand() & 0x1f))  ;
	}	
}

// Set all densities in a cell to their equilibrium values for a given velocity and density:
// (If density is omitted, it's left unchanged.)
void setEquil(int x, int y, float newux, float newuy, float newrho) {
	int i = x + y*xdim;
	//
	float ux3 = 3.0f * (newux) ;// 3 * newux;
	float uy3 = 3.0f * (newuy) ;// 3 * newuy;
	float ux2 = (newux * newux) ;
	float uy2 = (newuy * newuy);
	float uxuy2 = (newux * newuy) * 2 ; //2 * newux * newuy;
	float u2 = ux2 + uy2;
	float u215 = 1.5f * (u2) ; // 1.5 * u2;
    float one9th_newrho = (one9th * newrho) ;
    float one36th_newrho = (one36th * newrho) ;
    float four9ths_newrho = (four9ths * newrho) ;
	//n0[i]  = mFix(four9ths_newrho, (1 - u215));
	nE[i]  = ( one9th_newrho * (1.0f + ux3       + 4.5f*(ux2)      - u215)) ;
	nW[i]  = ( one9th_newrho * (1.0f - ux3       + 4.5f*(ux2)      - u215)) ;
	nN[i]  = ( one9th_newrho * (1.0f + uy3       + 4.5f*(uy2)      - u215)) ;
	nS[i]  = ( one9th_newrho * (1.0f - uy3       + 4.5f*(uy2)      - u215)) ;
	nNE[i] = (one36th_newrho * (1.0f + ux3 + uy3 + 4.5f*(u2+uxuy2) - u215)) ;
	nSE[i] = (one36th_newrho * (1.0f + ux3 - uy3 + 4.5f*(u2-uxuy2) - u215)) ;
	nNW[i] = (one36th_newrho * (1.0f - ux3 + uy3 + 4.5f*(u2-uxuy2) - u215)) ;
	nSW[i] = (one36th_newrho * (1.0f - ux3 - uy3 + 4.5f*(u2+uxuy2) - u215)) ;
    n0[i] = newrho - (nE[i]+nW[i]+nN[i]+nS[i]+nNE[i]+nSE[i]+nNW[i]+nSW[i]) ;
	rho[i] = newrho;
	ux[i] = newux;
	uy[i] = newuy;
}

// Function to initialize or re-initialize the fluid, based on speed slider setting:
void initFluid() {
	// Amazingly, if I nest the y loop inside the x loop, Firefox slows down by a factor of 20
	float u0 = fluid_speed;
	for (int y=0; y<ydim; y++) {
		for (int x=0; x<xdim; x++) {
			setEquil(x, y, u0, 0, 1.0f);
		}
	}
} // end initFluid

// Set the fluid floatiables at the boundaries, according to the current slider value:
void setBoundaries() {
	float u0 = fluid_speed;
	for (int x=0; x<xdim; x++) {
		setEquil(x, 0, u0, 0, 1.0f);
		setEquil(x, ydim-1, u0, 0, 1.0f);
	}
	for (int y=1; y<ydim-1; y++) {
		setEquil(0, y, u0, 0, 1.0f);
		setEquil(xdim-1, y, u0, 0, 1.0f);
	}
} // end setBoundaries()//

// Collide particles within each cell (here's the physics!):
// !! Core 0 !!
void collide(void) {
	float viscosity = fluid_viscosity;	// kinematic viscosity coefficient in natural units
	float omega = (1.0f / (3.0f*(viscosity) + 0.5f));		// reciprocal of relaxation time
	// use these to split between the two processors
	#define xbottom 1
	#define xtop xdim/2
	//for (int y=1; y<ydim-1; y++) {
	//	for (int x=xbottom; x<xtop; x++) {
	for (int x=xbottom; x<xtop; x++) {
		for (int y=1; y<ydim-1; y++) {
			int i = x + y*xdim;		// array index for this lattice site
			float thisrho = n0[i] + nN[i] + nS[i] + nE[i] + nW[i] + nNW[i] + nNE[i] + nSW[i] + nSE[i];
			rho[i] = thisrho;
			//>>>float rho_inv = 1.0 - (thisrho - 1.0) + ((thisrho - 1.0) * (thisrho - 1.0));
            float rho_inv = 1.0f / thisrho;
            // >>>
			//Fix thisux = divFix((nE[i] + nNE[i] + nSE[i] - nW[i] - nNW[i] - nSW[i]), thisrho);
			float thisux = ((nE[i] + nNE[i] + nSE[i] - nW[i] - nNW[i] - nSW[i]) * rho_inv);
			ux[i] = thisux;
			//Fix thisuy = divFix((nN[i] + nNE[i] + nNW[i] - nS[i] - nSE[i] - nSW[i]), thisrho);
			float thisuy = ((nN[i] + nNE[i] + nNW[i] - nS[i] - nSE[i] - nSW[i]) * rho_inv);
			uy[i] = thisuy ;
			float one9thrho = (one9th * thisrho);		// pre-compute a bunch of stuff for optimization
			float one36thrho = (one36th * thisrho);
			float ux3 = 3.0f * (thisux);
			float uy3 = 3.0f * (thisuy);
			float ux2 = (thisux * thisux);
			float uy2 = (thisuy * thisuy);
			float uxuy2 = 2.0f * (thisux * thisuy);
			float u2 = ux2 + uy2;
			float u215 = 1.5f * (u2);
            
			//n0[i]  += omega * (four9ths*thisrho * (1                        - u215) - n0[i]);
			nE[i]  += (omega * ( (one9thrho * (1.0f + ux3    +  4.5f*(ux2)      - u215)) - nE[i]));
			nW[i]  += (omega * ( (one9thrho * (1.0f - ux3    +  4.5f*(ux2)      - u215)) - nW[i]));
			nN[i]  += (omega * ( (one9thrho * (1.0f + uy3    +  4.5f*(uy2)      - u215)) - nN[i]));
			nS[i]  += (omega * ( (one9thrho * (1.0f - uy3    +  4.5f*(uy2)      - u215)) - nS[i]));
			nNE[i] += (omega * ( (one36thrho * (1.0f + ux3 + uy3 +  4.5f*(u2+uxuy2) - u215)) - nNE[i]));
			nSE[i] += (omega * ( (one36thrho * (1.0f + ux3 - uy3 +  4.5f*(u2-uxuy2) - u215)) - nSE[i]));
			nNW[i] += (omega * ( (one36thrho * (1.0f - ux3 + uy3 +  4.5f*(u2-uxuy2) - u215)) - nNW[i]));
			nSW[i] += (omega * ( (one36thrho * (1.0f - ux3 - uy3 +  4.5f*(u2+uxuy2) - u215)) - nSW[i]));
			// force conservation of mass for limited precision aithmetic
			n0[i]   = thisrho - (nE[i]+nW[i]+nN[i]+nS[i]+nNE[i]+nSE[i]+nNW[i]+nSW[i]);
		}
	}
    // dont need the following for 2 cores
	//for (int y=1; y<ydim-2; y++) {
	//	nW[xdim-1+y*xdim] = nW[xdim-2+y*xdim];		// at right end, copy left-flowing densities from next row to the left
	//	nNW[xdim-1+y*xdim] = nNW[xdim-2+y*xdim];
	//	nSW[xdim-1+y*xdim] = nSW[xdim-2+y*xdim];
	//}
} // end collide//

// Collide particles within each cell (here's the physics!):
// !! CORE 1 !!
void collide1(void) {
	float viscosity = fluid_viscosity;	// kinematic viscosity coefficient in natural units
	float omega = (1.0f / (3.0f*(viscosity) + 0.5f));		// reciprocal of relaxation time
	// use these to split between the two processors
	#define xbottom1 xdim/2
	#define xtop1 xdim
	for (int y=1; y<ydim-1; y++) {
		for (int x=xbottom1; x<xtop1-1; x++) {
			int i = x + y*xdim;		// array index for this lattice site
			float thisrho = n0[i] + nN[i] + nS[i] + nE[i] + nW[i] + nNW[i] + nNE[i] + nSW[i] + nSE[i];
			rho[i] = thisrho;
			//>>>float rho_inv = 1.0 - (thisrho - 1.0) + ((thisrho - 1.0) * (thisrho - 1.0));
            float rho_inv = 1.0f / thisrho;
            // >>>
			//Fix thisux = divFix((nE[i] + nNE[i] + nSE[i] - nW[i] - nNW[i] - nSW[i]), thisrho);
			float thisux = ((nE[i] + nNE[i] + nSE[i] - nW[i] - nNW[i] - nSW[i]) * rho_inv);
			ux[i] = thisux;
			//Fix thisuy = divFix((nN[i] + nNE[i] + nNW[i] - nS[i] - nSE[i] - nSW[i]), thisrho);
			float thisuy = ((nN[i] + nNE[i] + nNW[i] - nS[i] - nSE[i] - nSW[i]) * rho_inv);
			uy[i] = thisuy ;
			float one9thrho = (one9th * thisrho);		// pre-compute a bunch of stuff for optimization
			float one36thrho = (one36th * thisrho);
			float ux3 = 3 * (thisux);
			float uy3 = 3 * (thisuy);
			float ux2 = (thisux * thisux);
			float uy2 = (thisuy * thisuy);
			float uxuy2 = 2 * (thisux * thisuy);
			float u2 = ux2 + uy2;
			float u215 = 1.5 * (u2);
            
			//n0[i]  += omega * (four9ths*thisrho * (1                        - u215) - n0[i]);
			nE[i]  += (omega * ( (one9thrho * (1.0f + ux3    +  4.5f*(ux2)      - u215)) - nE[i]));
			nW[i]  += (omega * ( (one9thrho * (1.0f - ux3    +  4.5f*(ux2)      - u215)) - nW[i]));
			nN[i]  += (omega * ( (one9thrho * (1.0f + uy3    +  4.5f*(uy2)      - u215)) - nN[i]));
			nS[i]  += (omega * ( (one9thrho * (1.0f - uy3    +  4.5f*(uy2)      - u215)) - nS[i]));
			nNE[i] += (omega * ( (one36thrho * (1.0f + ux3 + uy3 +  4.5f*(u2+uxuy2) - u215)) - nNE[i]));
			nSE[i] += (omega * ( (one36thrho * (1.0f + ux3 - uy3 +  4.5f*(u2-uxuy2) - u215)) - nSE[i]));
			nNW[i] += (omega * ( (one36thrho * (1.0f - ux3 + uy3 +  4.5f*(u2-uxuy2) - u215)) - nNW[i]));
			nSW[i] += (omega * ( (one36thrho * (1.0f - ux3 - uy3 +  4.5f*(u2+uxuy2) - u215)) - nSW[i]));
			// force conservation of mass for limited precision aithmetic
			n0[i]   = thisrho - (nE[i]+nW[i]+nN[i]+nS[i]+nNE[i]+nSE[i]+nNW[i]+nSW[i]);
		}
	}
	for (int y=1; y<ydim-2; y++) {
		nW[xdim-1+y*xdim] = nW[xdim-2+y*xdim];		// at right end, copy left-flowing densities from next row to the left
		nNW[xdim-1+y*xdim] = nNW[xdim-2+y*xdim];
		nSW[xdim-1+y*xdim] = nSW[xdim-2+y*xdim];
	}
} // end collide1

// Move particles along their directions of motion:
// core 0
void stream(void) {
	barrierCount = 0; barrierxSum = 0; barrierySum = 0;
	for (int y=ydim-2; y>0; y--) {			// first start in NW corner...
		for (int x=1; x<xdim-1; x++) {
			nN[x+y*xdim] = nN[x+(y-1)*xdim];			// move the north-moving particles
			nNW[x+y*xdim] = nNW[x+1+(y-1)*xdim];		// and the northwest-moving particles
		}
	}
	for (int y=ydim-2; y>0; y--) {			// now start in NE corner...
		for (int x=xdim-2; x>0; x--) {
			nE[x+y*xdim] = nE[x-1+y*xdim];			// move the east-moving particles
			nNE[x+y*xdim] = nNE[x-1+(y-1)*xdim];		// and the northeast-moving particles
		}
	}
	for (int y=1; y<ydim-1; y++) {			// now start in SE corner...
		for (int x=xdim-2; x>0; x--) {
			nS[x+y*xdim] = nS[x+(y+1)*xdim];			// move the south-moving particles
			nSE[x+y*xdim] = nSE[x-1+(y+1)*xdim];		// and the southeast-moving particles
		}
	}
    /*
	for (int y=1; y<ydim-1; y++) {				// now start in the SW corner...
		for (int x=1; x<xdim-1; x++) {
			nW[x+y*xdim] = nW[x+1+y*xdim];			// move the west-moving particles
			nSW[x+y*xdim] = nSW[x+1+(y+1)*xdim];		// and the southwest-moving particles
		}
	}
	for (int y=1; y<ydim-1; y++) {				// Now handle bounce-back from barriers
		for (int x=1; x<xdim-1; x++) {
			if (barrier[x+y*xdim]) {
				int index = x + y*xdim;
				nE[x+1+y*xdim] = nW[index];
				nW[x-1+y*xdim] = nE[index];
				nN[x+(y+1)*xdim] = nS[index];
				nS[x+(y-1)*xdim] = nN[index];
				nNE[x+1+(y+1)*xdim] = nSW[index];
				nNW[x-1+(y+1)*xdim] = nSE[index];
				nSE[x+1+(y-1)*xdim] = nNW[index];
				nSW[x-1+(y-1)*xdim] = nNE[index];
			}
		}
	}
    */
} // end stream

// !! CORE 1 !!
// Move particles along their directions of motion:
void stream1(void) {
    /*
	barrierCount = 0; barrierxSum = 0; barrierySum = 0;
	for (int y=ydim-2; y>0; y--) {			// first start in NW corner...
		for (int x=1; x<xdim-1; x++) {
			nN[x+y*xdim] = nN[x+(y-1)*xdim];			// move the north-moving particles
			nNW[x+y*xdim] = nNW[x+1+(y-1)*xdim];		// and the northwest-moving particles
		}
	}
	for (int y=ydim-2; y>0; y--) {			// now start in NE corner...
		for (int x=xdim-2; x>0; x--) {
			nE[x+y*xdim] = nE[x-1+y*xdim];			// move the east-moving particles
			nNE[x+y*xdim] = nNE[x-1+(y-1)*xdim];		// and the northeast-moving particles
		}
	}
	for (int y=1; y<ydim-1; y++) {			// now start in SE corner...
		for (int x=xdim-2; x>0; x--) {
			nS[x+y*xdim] = nS[x+(y+1)*xdim];			// move the south-moving particles
			nSE[x+y*xdim] = nSE[x-1+(y+1)*xdim];		// and the southeast-moving particles
		}
	}
    */
	for (int y=1; y<ydim-1; y++) {				// now start in the SW corner...
		for (int x=1; x<xdim-1; x++) {
			nW[x+y*xdim] = nW[x+1+y*xdim];			// move the west-moving particles
			nSW[x+y*xdim] = nSW[x+1+(y+1)*xdim];		// and the southwest-moving particles
		}
	}
	for (int y=1; y<ydim-1; y++) {				// Now handle bounce-back from barriers
		for (int x=1; x<xdim-1; x++) {
			if (barrier[x+y*xdim]) {
				int index = x + y*xdim;
				nE[x+1+y*xdim] = nW[index];
				nW[x-1+y*xdim] = nE[index];
				nN[x+(y+1)*xdim] = nS[index];
				nS[x+(y-1)*xdim] = nN[index];
				nNE[x+1+(y+1)*xdim] = nSW[index];
				nNW[x-1+(y+1)*xdim] = nSE[index];
				nSE[x+1+(y-1)*xdim] = nNW[index];
				nSW[x-1+(y-1)*xdim] = nNE[index];
			}
		}
	}
} // end stream1
	
// Add a barrier at a given grid coordinate location:
void addBarrier(int x, int y) {
    if ((x > 1) && (x < xdim-2) && (y > 1) && (y < ydim-2)) {
        barrier[x+y*xdim] = true;
    }
}

// Remove a barrier at a given grid coordinate location:
void removeBarrier(int x, int y) {
    if (barrier[x+y*xdim]) {
        barrier[x+y*xdim] = false;
    }
}

// Clear all barriers:
void clearBarriers(void) {
    for (int x=0; x<xdim; x++) {
        for (int y=0; y<ydim; y++) {
            barrier[x+y*xdim] = false;
        }
    }
}

// ==================================================
// === Lattice Boltzmann demo -- RUNNING on core 0
// ==================================================


static PT_THREAD (protothread_graphics(struct pt *pt)) {
    PT_BEGIN(pt);
    
    // Draw some filled rectangles
    fillRect(0, 0, 165, 50, BLUE); // blue box
    fillRect(170, 0, 165, 50, YELLOW); // blue box

    // Write some text
    setTextColor(WHITE) ;
    setCursor(5, 0) ;
    setTextSize(1) ;
    writeString("RP2350 Pico2 floating Pt.") ;
    setCursor(5, 10) ;
    writeString("Lattice Boltzmann Flow") ;
    setCursor(5, 20) ;
    writeString("Hunter Adams / Bruce Land") ;
    setCursor(5, 30) ;
    writeString("brl4@cornell.edu") ;
	setCursor(5, 40) ;
    writeString("320x240 16 color VGA") ;

    // fillRect(short x, short y, short w, short h, char color) ;
    for(int i=0; i<16; i++){
        fillRect(i*10+70, 150, 10, 10, i) ;
    }
    setCursor(70, 165) ;
    writeString("slow") ;
    setCursor(70+150, 165) ;
    writeString("fast") ;

    char video_buffer[32];
    static uint64_t elapsed_time ;

    // the Lattice-boltzmaann setup
    init_LB();
    // initialize to steady rightward flow
    initFluid();	
    //
	init_trace();

    // put in a vertical barrier
	for (int y=ydim/2-5; y<ydim/2+4; y++) {
		addBarrier(15, y) ;
		addBarrier(16, y) ;
		addBarrier(17, y) ;
        //addBarrier(50, y+5) ;
	}

	// for (int y=ydim-5; y<ydim ; y++) {
	// 	addBarrier(100, y) ;	
    //     //addBarrier(50, y+5) ;
	// }
    // 

    while(true) {
      // A brief nap
      PT_YIELD(pt) ;

      // compute one time iteration 
      elapsed_time = PT_GET_TIME_usec();
      // main simulation loop 
	  setBoundaries();	
	  // Execute update of left half of region
      // tell core1 to run right half
      PT_SEM_SAFE_SIGNAL(pt, &collide_go_s) ;
	  collide();
      // wait for core 1 to finish
      PT_SEM_SAFE_WAIT(pt, &collide_done_s) ;
      // tell core 1 to stream
      PT_SEM_SAFE_SIGNAL(pt, &stream_go_s) ;
	  stream();
      PT_SEM_SAFE_WAIT(pt, &stream_done_s) ;

	 ////
	  move_trace();

        for (int x=0; x<xdim; x++) {
            for (int y=0; y<ydim; y++) {
                int i = x+y*xdim ;
                //short speed = 50*sqrt(ux[x+y*xdim]*ux[x+y*xdim]+uy[x+y*xdim]*uy[x+y*xdim]);
                // VERY approx mag estimate
                // shift 19 for 25 bit fixed
                short speed = 60*(MAX(fabs(ux[i]), fabs(uy[i])) + 0.4f * (MIN(fabs(ux[i]), fabs(uy[i]))) ) ; //
                drawPixel(x+50, y+70, speed) ;				
            }
        }

        int i;
        for(i=0; i<n_trace; i++){
			drawPixel(50+(int)(trace_x[i]), 70+(int)(trace_y[i]),  BLACK) ;
		}

      elapsed_time = PT_GET_TIME_usec()-elapsed_time ;
      sprintf(video_buffer, "Grid size=%dx%d ", xdim, ydim);
      setCursor(175, 10) ;
      setTextColor2(BLACK, YELLOW) ;
      setTextSize(1) ;
      writeString(video_buffer) ;

      sprintf(video_buffer, "fluid speed=%4.3f  ", (fluid_speed));
      setCursor(175, 20) ;
      setTextColor2(BLACK, YELLOW) ;
      setTextSize(1) ;
      writeString(video_buffer) ;

      sprintf(video_buffer, "viscosity=%4.3f ",  (fluid_viscosity));
      setCursor(175, 30) ;
      setTextColor2(BLACK, YELLOW) ;
      setTextSize(1) ;
      writeString(video_buffer) ;

      sprintf(video_buffer, "Time/frame=%d mSec ", elapsed_time/1000);
      setCursor(175, 40) ;
      setTextColor2(BLACK, YELLOW) ;
      setTextSize(1) ;
      writeString(video_buffer) ;
//
   }

   PT_END(pt);
} // graphics thread

// ==================================================
// === toggle25 thread on core 0
// ==================================================
// the on-board LED blinks
static PT_THREAD (protothread_toggle25(struct pt *pt))
{
    PT_BEGIN(pt);
    static bool LED_state = false ;
    
     // set up LED p25 to blink
     gpio_init(25) ;	
     gpio_set_dir(25, GPIO_OUT) ;
     gpio_put(25, true);
     // data structure for interval timer
     PT_INTERVAL_INIT() ;

      while(1) {
        // yield time 0.1 second
        //PT_YIELD_usec(100000) ;
        PT_YIELD_INTERVAL(200000) ;

        // toggle the LED on PICO
        LED_state = LED_state? false : true ;
        gpio_put(25, LED_state);
        //
        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // blink thread

// ==================================================
// === LB collide -- RUNNING on core 1
// ==================================================
// Lattice Boltamann collide thead
static PT_THREAD (protothread_LB_collide(struct pt *pt))
{
    PT_BEGIN(pt);
    
      while(1) {
        //
        PT_YIELD(pt) ;
        // Execute update of right half of region
        // tell core1 to run
        PT_SEM_SAFE_WAIT(pt, &collide_go_s) ;
        collide1();
        // tell core 0 it is done
        PT_SEM_SAFE_SIGNAL(pt, &collide_done_s) ;

        PT_SEM_SAFE_WAIT(pt, &stream_go_s) ;
	    stream1();
         PT_SEM_SAFE_SIGNAL(pt, &stream_done_s) ;
        
        // NEVER exit while
      } // END WHILE(1)
    PT_END(pt);
} // blink thread

// ==================================================
// === user's serial input thread on core 1
// ==================================================
// serial_read an serial_write do not block any thread
// except this one
static PT_THREAD (protothread_serial(struct pt *pt))
{
    PT_BEGIN(pt);
      static int test_in1, test_in2, sum ;
      //
      while(1) {
        // print prompt
        sprintf(pt_serial_out_buffer, "input x y: ");
        // spawn a thread to do the non-blocking write
        serial_write ;

        // spawn a thread to do the non-blocking serial read
         serial_read ;
        
        // convert input string to number
        //sscanf(pt_serial_in_buffer,"%d %d ",  &col1, &col2) ;
        //printf("color=%d\n\r", readPixel(col1, col2)) ;

        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // serial thread

// ========================================
// === core 1 main -- started in main below
// ========================================
void core1_main(){ 
  //
  //  === add threads  ====================
  // for core 1
  pt_add_thread(protothread_LB_collide) ;
  //pt_add_thread(protothread_serial) ;
  //
  // === initalize the scheduler ==========
  pt_schedule_start ;
  // NEVER exits
  // ======================================
}

// ========================================
// === core 0 main
// ========================================
int main(){
  // set the clock
 // set_sys_clock_khz(300000, true); // 171us

  sleep_ms(20) ;//
  // start the serial i/o
  stdio_init_all() ;
  // announce the threader version on system reset
  printf("\n\rProtothreads RP2350 v1.3 two-core\n\r");


  // Initialize the VGA screen
  initVGA() ;

  // four safe semaphores
   PT_SEM_SAFE_INIT(&collide_go_s, 0) ;
   PT_SEM_SAFE_INIT(&collide_done_s, 0) ;
   PT_SEM_SAFE_INIT(&stream_go_s, 0) ;
   PT_SEM_SAFE_INIT(&stream_done_s, 0) ;
     
  // start core 1 threads
  //multicore_reset_core1();
  multicore_launch_core1(&core1_main);
  sleep_ms(20);

  // === config threads ========================
  // for core 0
  pt_add_thread(protothread_graphics);
  pt_add_thread(protothread_toggle25);
  //
  // === initalize the scheduler ===============
  pt_schedule_start ;
  // NEVER exits
  // ===========================================
} // end main